# Offline First Database Comparison

In this project I have implemented the exact same chat application with different database technologies.
You can use it to compare metrics and learn about the differences. The chat app is a web based **angular** application, with functionality similar to Whatsapp Web.

<p align="center">
  <img src="./orga/images/chat-app.png" alt="chat app" width="450" />
</p>

## Implemented Databases:
  - AWS Amplify Datastore
  - Firebase Firestore
  - [PouchDB](https://github.com/pouchdb/pouchdb) with IndexedDB adapter & CouchDB replication
  - [RxDB](https://github.com/pubkey/rxdb) with PouchDB Storage & GraphQL replication
  - [RxDB](https://github.com/pubkey/rxdb) with LokiJS Storage & GraphQL replication
  - [WatermelonDB](https://github.com/Nozbe/WatermelonDB) with LokiJS adapter (no backend sync atm)

## Metrics

All metrics are measured automatically via code in a browser tests (chrome:headless). The results heavily depend on the developers device. You should compare the values relative to another and not as absolute values. Also you might want to create new metrics that better represent how you would use the respective database.

You can reproduce these values by running `sh measure-metrics.sh` in the root folder.

<!-- !!! DO NOT EDIT THIS TABLE DIRECTLY, IT IS GENERATED BY ./scripts/aggregate-metrics.mjs !!! -->

| Metric \ Project                        | aws    | firebase | pouchdb | rxdb-lokijs | rxdb-pouchdb | watermelondb |
| --------------------------------------- | ------ | -------- | ------- | ----------- | ------------ | ------------ |
| First angular component render          | 289ms  | 327ms    | 264ms   | 207ms       | 215ms        | 192ms        |
| Page load time                          | 382ms  | 246ms    | 345ms   | 290ms       | 298ms        | 268ms        |
| First full render                       | 560ms  | 1648ms   | 1585ms  | 647ms       | 1528ms       | 315ms        |
| Insert one message                      | 47ms   | 432ms    | 27ms    | 13ms        | 26ms         | 7ms          |
| Inserting 20 messages one after another | 747ms  | 9795ms   | 681ms   | 1350ms      | 1829ms       | 32ms         |
| Inserting 20 messages in parallel       | 340ms  | 7982ms   | 146ms   | 1080ms      | 1718ms       | 1547ms       |
| Message insert to message list change   | 89ms   | 32ms     | 450ms   | 13ms        | 25ms         | 5ms          |
| User change to message list change      | 22ms   | 194ms    | 349ms   | 5ms         | 316ms        | 5ms          |
| Message search query time               | 600ms  | 408ms    | 355ms   | 27ms        | 124ms        | 28ms         |
| First full render with many messages    | 493ms  | 2308ms   | 2012ms  | 623ms       | 2589ms       | 293ms        |
| Storage usage                           | 200kb  | 3250kb   | 1369kb  | 228kb       | 2289kb       | 1767kb       |
| Bundle size, plain JavaScript           | 1573kb | 915kb    | 801kb   | 1083kb      | 1082kb       | 852kb        |
| Bundle size, minified+gzip              | 364kb  | 224kb    | 192kb   | 269kb       | 277kb        | 199kb        |


### Metrics Explanation
- Page load time: How long does it take to download and parse the JavaScript bundle.
- First angular component render: How long does it take for the first angular component to be rendered.
- First full render: How long does it take until all relevant data is displayed for the first time.
- Insert one message: How long does it take to insert a single message.
- Inserting 20 messages one after another: How long does it take to insert 20 messages in serial.
- Inserting 20 messages in parallel: How long does it take to insert 20 messages in parallel.
- Message insert to message list change: How long does it take until a new message is rendered to the dom.
- User change to message list change: How long does it take from changing the user to the displaying of the new messages list.
- Message search query time: How long does it take to search for a given message by regex/like-operator.
- First full render with many messages: Time to first full render when many messages exist. 
- Storage usage: Size of the stored IndexedDB database after inserting the full test dataset.
- Bundle size, plain JavaScript: The full JavaScript bundle size, without minification or gzip.
- Bundle size, minified+gzip: The full JavaScript bundle size after minification and gzip compression.

## Investigations

### Why is LokiJS so much faster?

WatermelonDB and the RxDB-LokiJS project use the [LokiJS](https://github.com/techfort/LokiJS) database as storage, which is an **in memory** database that regularly persists the data to IndexedDB either on interval, or when the browser tab is closed. Keeping and processing the data in memory has the benefit of being much faster, but it also has its downsides:

- Data can be lost when the JavaScript process is killed ungracefully like when the browser crashes or the power of the PC is terminated.
- There is no multi-tab-support. The data is not shared between multiple browser tabs of the same origin.
- There is no concept of conflict handling or transactions. The last write always wins.


### Why is Firebase so slow on first render?

On the first page load, Firebase ensures that the local data is equal to the server side state. This means that the client has to be online at application startup which is the reason why Firebase is not completely offline first. To ensure the equalness of client side data, Firebase has to perform several requests to the backend, before the database will respond to queries. This makes the inital page load slow, and it becomes even more slower, the more data exists and has to be validated.

### Why is PouchDB so slow?

For the PouchDB and RxDB (with PouchDB storage) I used the [old](https://www.npmjs.com/package/pouchdb-adapter-idb) Indexeddb adapter.
It is much less optimized than the [new](https://www.npmjs.com/package/pouchdb-adapter-indexeddb) adapter, but the new one made problems with returning the correct query results.
Theses problems have been fixed on the PouchDB master branch, but I have to wait for the next PouchDB release. I will update the repo when this change can be done.

### Why does AWS Datastore need so much less storage space?

AWS Datastore does not save any metadata together with the documents. Instead only the plain documents are stored in IndexedDB. They can do this because they only allow simple queries and do not keep a local version history.

## Feature Map

<!-- !!! DO NOT EDIT THIS TABLE DIRECTLY, IT IS GENERATED BY ./scripts/generate-feature-table.ts !!! -->

| Feature \ Project        | aws                | firebase                                     | pouchdb | rxdb-pouchdb | rxdb-lokijs | watermelondb                                                                             |
| ------------------------ | ------------------ | -------------------------------------------- | ------- | ------------ | ----------- | ---------------------------------------------------------------------------------------- |
| Offline First            | No, login required | Partially, must be online on first page load | **Yes** | **Yes**      | **Yes**     | **Yes**                                                                                  |
| Realtime Replication     | **Yes**            | **Yes**                                      | **Yes** | **Yes**      | **Yes**     | Partially, must be implemented by hand                                                   |
| Multi Tab Support        | **Yes**            | **Yes**                                      | No      | **Yes**      | **Yes**     | Partially, relies on online sync                                                         |
| Observable Queries       | No                 | **Yes**                                      | No      | **Yes**      | **Yes**     | **Yes**                                                                                  |
| Complex Queries          | No                 | **Yes**                                      | **Yes** | **Yes**      | **Yes**     | Partially, limit/skip/sort not working with LokiJS adapter (fixed in latest pre-release) |
| Client Side Encryption   | No                 | No                                           | **Yes** | **Yes**      | **Yes**     | No                                                                                       |
| Schema Support           | **Yes**            | No                                           | No      | **Yes**      | **Yes**     | **Yes**                                                                                  |
| Custom Backend           | No                 | No                                           | No      | **Yes**      | **Yes**     | **Yes**                                                                                  |
| Custom Conflict Handling | **Yes**            | No                                           | **Yes** | **Yes**      | No          | No                                                                                       |


## Starting the projects

All sub-projects use the same port and so can **not be started in parallel**.

### Installation

* You must have [installed Node.js](https://nodejs.org/en/download/)
* Clone this project
* In the root folder, run `npm install` to install the dependencies.
* In the root folder, run `npm run build` to build all projects.


### Firebase Firestore

* Run `npm run start:firebase` to start the mock server and the production build frontend.
* Or run `npm run dev:firebase` to start the mock server and the development frontend server.

* Open [http://localhost:3000/](http://localhost:3000/) to browse the frontend.

### AWS Amplify & Datastore

The official AWS mock does [not allow a live replication](https://github.com/aws-amplify/amplify-cli/issues/4155) at this point. So you first have to setup an amplify project in the `./projects/aws` folder by using [this tutorial](https://docs.amplify.aws/lib/datastore/getting-started/q/platform/js/)

* Run `npm run start:aws` to start the mock server and the production build frontend.
* Or run `npm run dev:aws` to start the mock server and the development frontend server.

* Open [http://localhost:3000/](http://localhost:3000/) to browse the frontend.

### PouchDB

* Run `npm run start:pouchdb` to start the mock server and the production build frontend.
* Or run `npm run dev:pouchdb` to start the mock server and the development frontend server.

* Open [http://localhost:3000/](http://localhost:3000/) to browse the frontend.

### RxDB

* Run `npm run start:rxdb` to start the mock server and the production build frontend.
* Or run `npm run dev:rxdb` to start the mock server and the development frontend server.

* Open [http://localhost:3000/](http://localhost:3000/) to browse the frontend.


### WatermelonDB

* Run `npm run start:watermelondb` to start the mock server and the production build frontend.
* Or run `npm run dev:watermelondb` to start the mock server and the development frontend server.

* Open [http://localhost:3000/](http://localhost:3000/) to browse the frontend.

## TODOs

Pull requests are welcomed. Please help implementing more examples:

- Meteor (with the IndexedDB offline first plugin).
- WatermelonDB backend replication.
- AWS Ampflify local backend mock with realtime replication.
- GunDB.
